import io
from pathlib import Path
from typing import Dict, Optional, Sequence, Literal, Any, Union, Callable

import pandas as pd
from pydantic import BaseModel, Field


class Style(BaseModel):
    class Config:
        frozen = True

    align: Optional[
        Literal[
            "left", "center", "right", "fill", "justify", "center_across", "distributed"
        ]
    ] = Field(default=None)
    background: Optional[str] = Field(default=None)
    bold: Optional[bool] = Field(default=None)
    border: Optional[int] = Field(default=None)
    border_bottom: Optional[int] = Field(default=None)
    border_color: Optional[str] = Field(default=None)
    border_left: Optional[int] = Field(default=None)
    border_right: Optional[int] = Field(default=None)
    border_top: Optional[int] = Field(default=None)
    fill_inf: Optional[Any] = Field(default=None)
    fill_na: Optional[Any] = Field(default=None)
    fill_zero: Optional[str] = Field(default=None)
    font_color: Optional[str] = Field(default=None)
    font_family: Optional[str] = Field(default=None)
    font_size: Optional[int] = Field(default=None)
    numeric_format: Optional[str] = Field(default=None)
    padding: Optional[int] = Field(default=None)
    padding_bottom: Optional[int] = Field(default=None)
    padding_left: Optional[int] = Field(default=None)
    padding_right: Optional[int] = Field(default=None)
    padding_top: Optional[int] = Field(default=None)
    text_wrap: Optional[bool] = Field(default=None)
    underline: Optional[Literal[1, 2, 33, 34]] = Field(default=None)
    valign: Optional[
        Literal["top", "vcenter", "bottom", "vcenter", "bottom", "vjustify"]
    ] = Field(default=None)

    def merge(self, other: "Style") -> "Style":
        self_dict = self.model_dump(exclude_none=True)
        other_dict = other.model_dump(exclude_none=True)
        self_dict.update(other_dict)
        return self.model_validate(self_dict)

    def pl(self) -> int:
        return self.padding_left or self.padding or 0

    def pt(self) -> int:
        return self.padding_top or self.padding or 0

    def pr(self) -> int:
        return self.padding_right or self.padding or 0

    def pb(self) -> int:
        return self.padding_bottom or self.padding or 0


class Component(BaseModel):
    style: Style = Field(default_factory=Style)

    class Config:
        arbitrary_types_allowed = True


class Text(Component):
    text: str
    width: int = Field(default=1)
    height: int = Field(default=1)
    merged: bool = Field(default=True)


class Link(Component):
    text: str
    url: str
    width: int = Field(default=1)
    height: int = Field(default=1)
    merged: bool = Field(default=True)

    def __str__(self):
        return self.text


class Fill(Component):
    width: int = Field(default=1)
    height: int = Field(default=1)
    merged: bool = Field(default=True)


class Image(Component):
    path: Path
    width: int = Field(default=1)
    height: int = Field(default=1)


class Table(Component):
    data: pd.DataFrame
    header_style: Dict[str, Style] = Field(default_factory=dict)
    body_style: Style = Field(default_factory=Style)
    column_style: Dict[str, Union[Style, Callable[[Any], Style]]] = Field(default_factory=dict)
    idx_column_style: Dict[int, Union[Style, Callable[[Any], Style]]] = Field(default_factory=dict)
    column_width: Dict[str, int] = Field(default_factory=dict)
    idx_column_width: Dict[int, int] = Field(default_factory=dict)
    row_style: Dict[int, Style] = Field(default_factory=dict)
    max_col_width: Optional[int] = Field(default=None)
    header_filters: bool = Field(default=True)
    default_style: bool = Field(default=True)
    auto_width_tuning: int = Field(default=5)
    auto_width_padding: int = Field(default=5)
    merge_equal_headers: bool = Field(default=True)

    def with_stripes(
        self,
        color: str = "#D0D0D0",
        pattern: Literal["even", "odd"] = "odd",
    ) -> "Table":
        return self.model_copy(
            update=dict(
                row_style={
                    idx: (
                        self.row_style.get(idx, Style()).merge(Style(background=color))
                        if (pattern == "odd" and idx % 2 != 0)
                        or (pattern == "even" and idx % 2 == 0)
                        else self.row_style.get(idx, Style())
                    )
                    for idx in range(self.data.shape[0])
                }
            )
        )


class Sheet(BaseModel):
    name: str
    components: Sequence[Component] = Field(default_factory=list)
    grid_lines: bool = Field(default=True)
    style: Style = Field(default_factory=Style)


class Excel(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    path: Union[Path, io.BytesIO]
    sheets: Sequence[Sheet] = Field(default_factory=list)
    nan_inf_to_errors: bool = Field(default=True)



def python_to_excel_fmt(fmt: str) -> str:
    fmt = fmt.strip()
    if fmt.endswith("%"):
        decimals = int(fmt.strip("%").lstrip(".") or 0)
        return f"0.{decimals * '0'}%" if decimals > 0 else "0%"
    if "," in fmt and "f" in fmt:
        decimals = int(fmt.split(".")[-1].rstrip("f") or 0)
        return f"#,##0.{decimals * '0'}" if decimals > 0 else "#,##0"
    if "f" in fmt:
        decimals = int(fmt.split(".")[-1].rstrip("f") or 0)
        return f"0.{decimals * '0'}" if decimals > 0 else "0"
    return "General"


PROP_MAP = dict(
    align="align",
    valign="valign",
    font_size="font_size",
    font_color="font_color",
    font_family="font_name",
    bold="bold",
    border="border",
    border_left="left",
    border_right="right",
    border_top="top",
    border_bottom="bottom",
    border_color="border_color",
    background="bg_color",
    numeric_format="num_format",
    underline="underline",
)

PRE_PROCESS_MAP = dict(
    numeric_format=python_to_excel_fmt,
)



import logging
from typing import Tuple, Dict, Callable

import xlsxwriter
from xlsxwriter.workbook import Workbook, Worksheet

from excelipy.models import Component, Excel, Fill, Style, Table, Text, Image, Link
from excelipy.writers import (
    write_fill,
    write_table,
    write_text,
    write_image, write_link,
)

log = logging.getLogger("excelipy")


def write_component(
        workbook: Workbook,
        worksheet: Worksheet,
        component: Component,
        default_style: Style,
        origin: Tuple[int, int] = (0, 0),
) -> Tuple[int, int]:
    writing_map: Dict[
        Callable[..., Component],
        Callable[..., Tuple[int, int]]
    ] = {
        Table: write_table,
        Text: write_text,
        Link: write_link,
        Fill: write_fill,
        Image: write_image,
    }

    render_func = writing_map.get(type(component))

    return render_func(
        workbook,
        worksheet,
        component,
        default_style,
        origin,
    )


def save(excel: Excel):
    with xlsxwriter.Workbook(excel.path, {
        "nan_inf_to_errors": excel.nan_inf_to_errors,
    }) as workbook:
        for sheet in excel.sheets:
            origin = (
                sheet.style.pl(),
                sheet.style.pt(),
            )
            worksheet = workbook.add_worksheet(sheet.name)
            if not sheet.grid_lines:
                worksheet.hide_gridlines(2)

            for component in sheet.components:
                cur_origin = (
                    origin[0] + component.style.pl(),
                    origin[1] + component.style.pt(),
                )
                x, y = write_component(
                    workbook,
                    worksheet,
                    component,
                    sheet.style,
                    cur_origin,
                )
                origin = origin[0] + component.style.pr(), origin[1] + y + component.style.pb()
